/* BFD back-end for Intel Hex objects.
   Copyright 1995, 1996, 1997, 1998, 1999 Free Software Foundation, Inc.
   Written by Ian Lance Taylor of Cygnus Support <ian@cygnus.com>.

This file is part of BFD, the Binary File Descriptor library.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*
  This module is designed to output assembly language for the MSP430 processor.
  It has no input capability.  It was built out of ihex.c.
*/

#include "bfd.h"
#include "sysdep.h"
#include "libbfd.h"
#include "libiberty.h"

#include <ctype.h>

static void msp430asm_input_error PARAMS ((bfd *));
static boolean msp430asm_mkobject PARAMS ((bfd *));
static const bfd_target *msp430asm_object_p PARAMS ((bfd *));
static boolean msp430asm_get_section_contents
  PARAMS ((bfd *, asection *, PTR, file_ptr, bfd_size_type));
static boolean msp430asm_set_section_contents
  PARAMS ((bfd *, asection *, PTR, file_ptr, bfd_size_type));
static boolean msp430asm_write_text_line PARAMS ((bfd *,const char *));
static boolean msp430asm_check_duplicate_symbol(const char *,asymbol **);
static boolean msp430asm_write_object_contents PARAMS ((bfd *));
static asymbol *msp430asm_make_empty_symbol PARAMS ((bfd *));
static boolean msp430asm_set_arch_mach
  PARAMS ((bfd *, enum bfd_architecture, unsigned long));
static int msp430asm_sizeof_headers PARAMS ((bfd *, boolean));

/* The maximum length of a string */
#define MAX_STRING 200

/* The number of bytes we put on one line during output.  */

#define CHUNK 8

/* When we write out an msp430asm value, the values can not be output as
   they are seen.  Instead, we hold them in memory in this structure.  */

struct msp430asm_data_list
{
  struct msp430asm_data_list *next;
  bfd_byte *data;
  bfd_vma where;
  bfd_size_type size;
	bfd mybfd;
	asection mysection;
	
};

/* The msp430asm tdata information.  */

struct msp430asm_data_struct
{
  struct msp430asm_data_list *head;
  struct msp430asm_data_list *tail;
};

/* Create an msp430asm object.  */

static boolean
msp430asm_mkobject (abfd)
	bfd *abfd;	
{
  if (abfd->tdata.msp430asm_data == NULL)
    {
      struct msp430asm_data_struct *tdata;

      tdata = ((struct msp430asm_data_struct *)
	       bfd_alloc (abfd, sizeof (struct msp430asm_data_struct)));
      if (tdata == NULL)
	return false;
      abfd->tdata.msp430asm_data = tdata;
      tdata->head = NULL;
      tdata->tail = NULL;
    }

  return true;
}


void msp430asm_input_error(abfd)
	bfd *abfd;
{
	(*_bfd_error_handler)
		(_("%s: Msp430asm target has no input capability\n"),
		 bfd_get_filename (abfd));
	bfd_set_error (bfd_error_bad_value);
}


/* No input capability, so we always return NULL */

static const bfd_target *
msp430asm_object_p (abfd)
     bfd *abfd;
{
	msp430asm_input_error(abfd);
	return NULL;
}

/* No input, so just error */

static boolean
msp430asm_get_section_contents (abfd, section, location, offset, count)
     bfd *abfd ATTRIBUTE_UNUSED;
     asection *section ATTRIBUTE_UNUSED;
     PTR location ATTRIBUTE_UNUSED;
     file_ptr offset ATTRIBUTE_UNUSED;
     bfd_size_type count ATTRIBUTE_UNUSED;
{
	msp430asm_input_error(abfd);
	return false;
}

/* Set the contents of a section in an Intel Hex file.  */

static boolean
msp430asm_set_section_contents (abfd, section, location, offset, count)
     bfd *abfd;
     asection *section;
     PTR location;
     file_ptr offset;
     bfd_size_type count;
{
  struct msp430asm_data_list *n;
  bfd_byte *data;
  struct msp430asm_data_struct *tdata;

  if (count == 0
      || (section->flags & SEC_ALLOC) == 0
      || (section->flags & SEC_LOAD) == 0)
    return true;

  n = ((struct msp430asm_data_list *)
       bfd_alloc (abfd, sizeof (struct msp430asm_data_list)));
  if (n == NULL)
    return false;

  data = (bfd_byte *) bfd_alloc (abfd, count);
  if (data == NULL)
    return false;
  memcpy (data, location, (size_t) count);

  n->data = data;
  n->where = section->lma + offset;
  n->size = count;
  n->mybfd = *abfd;
  n->mysection = *section;

  /* Sort the records by address.  Optimize for the common case of
     adding a record to the end of the list.  */
  tdata = abfd->tdata.msp430asm_data;
  if (tdata->tail != NULL
      && n->where >= tdata->tail->where)
    {
      tdata->tail->next = n;
      n->next = NULL;
      tdata->tail = n;
    }
  else
    {
      register struct msp430asm_data_list **pp;

      for (pp = &tdata->head;
	   *pp != NULL && (*pp)->where < n->where;
	   pp = &(*pp)->next)
	;
      n->next = *pp;
      *pp = n;
      if (n->next == NULL)
	tdata->tail = n;
    }

  return true;
}

/* Write out an Intel Hex file.  */

static boolean
msp430asm_write_text_line(abfd,line)
bfd *abfd;
const char *line;
{
  unsigned len;
  len = strlen(line);
  if (bfd_write (line, 1, len, abfd) != len)
	  return false;

  return true;
}

/* This is really inefficient */
static boolean
msp430asm_check_duplicate_symbol(sym,p)
const char *sym;
asymbol **p;
{
	for (; *p; p++){
		if(!strcmp(sym,(*p)->name))
			return(true);
	}
	return(false);
}
	
	

static boolean
msp430asm_write_object_contents (abfd)
     bfd *abfd;
{
  asymbol **p;
  struct msp430asm_data_list *l;
  char line[MAX_STRING+1];
  unsigned count,bc;
  static char bad_chars[]="\\/+-*=. ";	// List of chars that will make a bad msp430 symbol

  for (l = abfd->tdata.msp430asm_data->head; l != NULL; l = l->next)
    {
		sprintf(line,"\n\tASEG 0x%lx\t; Section ",l->where);
		if(!msp430asm_write_text_line(abfd,line))
			return(false);
		if(!msp430asm_write_text_line(abfd,l->mysection.name))
			return(false);
		if(!msp430asm_write_text_line(abfd,"\n"))
			return(false);

		for(count=0,bc=0;count<l->size;++count){
			if(!bc){
				sprintf(line,"\tDB 0x%02x",l->data[count]);
			}
			else{
				sprintf(line,",0x%02x",l->data[count]);
			}
			if(!msp430asm_write_text_line(abfd,line))
				return(false);

			if(++bc >= CHUNK){
				if(!msp430asm_write_text_line(abfd,"\n"))
					return(false);
				bc = 0;
			}
		}

		if(bc){
			if(!msp430asm_write_text_line(abfd,"\n"))
				return(false);
		}
	}
  
  /* And the symbols */
  if (abfd->outsymbols)
  {
	  if(!msp430asm_write_text_line(abfd,"\n\n\t;Symbols\n\n"))
		  return(false);
	  
      for (p = abfd->outsymbols; *p; p++)
	  {
		  int section_code = bfd_decode_symclass (*p);

		  if ((section_code != '?') && (section_code != 'U'))
		  {			/* do not include debug or unknown? symbols */
			  asymbol *s = *p;

			  if((!s->name) || (!s->name[0]))
				  continue;

			  if(isdigit(s->name[0]) || (strpbrk(s->name,bad_chars))){
				  if(!msp430asm_write_text_line(abfd,"; Invalid symbol "))
					  return(false);
				  
				  if(!msp430asm_write_text_line(abfd,s->name))
					  return(false);
				  
				  if(!msp430asm_write_text_line(abfd,"\n"))
					  return(false);

				  continue;
			  }
			  
			  if(msp430asm_check_duplicate_symbol(s->name,p+1))
				  continue;

			  if(!msp430asm_write_text_line(abfd,s->name))
				  return(false);
		  
			  if(!msp430asm_write_text_line(abfd,"\tEQU\t"))
				  return(false);

			  sprintf(line,"0x%lx\t; Section Code '%c'\n",s->value + s->section->vma,section_code);
			  if(!msp430asm_write_text_line(abfd,line))
				  return(false);
			  
			  if(!msp430asm_write_text_line(abfd,"\tEXPORT\t"))
				  return(false);
			  if(!msp430asm_write_text_line(abfd,s->name))
				  return(false);
			  if(!msp430asm_write_text_line(abfd,"\n"))
				  return(false);
		  
		  }
	  }
  }

  if(!msp430asm_write_text_line(abfd,"\n\tEND\n"))
	  return(false);
  
  return true;
}

/* Make an empty symbol.  This is required only because
   bfd_make_section_anyway wants to create a symbol for the section.  */

static asymbol *
msp430asm_make_empty_symbol (abfd)
     bfd *abfd;
{
  asymbol *new;

  new = (asymbol *) bfd_zalloc (abfd, sizeof (asymbol));
  if (new != NULL)
    new->the_bfd = abfd;
  return new;
}

/* Set the architecture for the output file.  The architecture is
   irrelevant, so we ignore errors about unknown architectures.  */

static boolean
msp430asm_set_arch_mach (abfd, arch, mach)
     bfd *abfd;
     enum bfd_architecture arch;
     unsigned long mach;
{
  if (! bfd_default_set_arch_mach (abfd, arch, mach))
    {
      if (arch != bfd_arch_unknown)
	return false;
    }
  return true;
}

/* Get the size of the headers, for the linker.  */

/*ARGSUSED*/
static int
msp430asm_sizeof_headers (abfd, exec)
     bfd *abfd ATTRIBUTE_UNUSED;
     boolean exec ATTRIBUTE_UNUSED;
{
  return 0;
}

/* Some random definitions for the target vector.  */

#define	msp430asm_close_and_cleanup _bfd_generic_close_and_cleanup
#define msp430asm_bfd_free_cached_info _bfd_generic_bfd_free_cached_info
#define msp430asm_new_section_hook _bfd_generic_new_section_hook
#define msp430asm_get_section_contents_in_window \
  _bfd_generic_get_section_contents_in_window

#define msp430asm_get_symtab_upper_bound bfd_0l
#define msp430asm_get_symtab \
  ((long (*) PARAMS ((bfd *, asymbol **))) bfd_0l)
#define msp430asm_print_symbol _bfd_nosymbols_print_symbol
#define msp430asm_get_symbol_info _bfd_nosymbols_get_symbol_info
#define msp430asm_bfd_is_local_label_name _bfd_nosymbols_bfd_is_local_label_name
#define msp430asm_get_lineno _bfd_nosymbols_get_lineno
#define msp430asm_find_nearest_line _bfd_nosymbols_find_nearest_line
#define msp430asm_bfd_make_debug_symbol _bfd_nosymbols_bfd_make_debug_symbol
#define msp430asm_read_minisymbols _bfd_nosymbols_read_minisymbols
#define msp430asm_minisymbol_to_symbol _bfd_nosymbols_minisymbol_to_symbol

#define msp430asm_get_reloc_upper_bound \
  ((long (*) PARAMS ((bfd *, asection *))) bfd_0l)
#define msp430asm_canonicalize_reloc \
  ((long (*) PARAMS ((bfd *, asection *, arelent **, asymbol **))) bfd_0l)
#define msp430asm_bfd_reloc_type_lookup _bfd_norelocs_bfd_reloc_type_lookup

#define msp430asm_bfd_get_relocated_section_contents \
  bfd_generic_get_relocated_section_contents
#define msp430asm_bfd_relax_section bfd_generic_relax_section
#define msp430asm_bfd_gc_sections bfd_generic_gc_sections
#define msp430asm_bfd_link_hash_table_create _bfd_generic_link_hash_table_create
#define msp430asm_bfd_link_add_symbols _bfd_generic_link_add_symbols
#define msp430asm_bfd_final_link _bfd_generic_final_link
#define msp430asm_bfd_link_split_section _bfd_generic_link_split_section

/* The Msp430asm target vector.  */

const bfd_target msp430asm_vec =
{
  "msp430asm",			/* name */
  bfd_target_msp430asm_flavour,
  BFD_ENDIAN_UNKNOWN,		/* target byte order */
  BFD_ENDIAN_UNKNOWN,		/* target headers byte order */
  0,				/* object flags */
  (SEC_HAS_CONTENTS | SEC_ALLOC | SEC_LOAD),	/* section flags */
  0,				/* leading underscore */
  ' ',				/* ar_pad_char */
  16,				/* ar_max_namelen */
  bfd_getb64, bfd_getb_signed_64, bfd_putb64,
  bfd_getb32, bfd_getb_signed_32, bfd_putb32,
  bfd_getb16, bfd_getb_signed_16, bfd_putb16,	/* data */
  bfd_getb64, bfd_getb_signed_64, bfd_putb64,
  bfd_getb32, bfd_getb_signed_32, bfd_putb32,
  bfd_getb16, bfd_getb_signed_16, bfd_putb16,	/* hdrs */

  {
    _bfd_dummy_target,
    msp430asm_object_p,		/* bfd_check_format */
    _bfd_dummy_target,
    _bfd_dummy_target,
  },
  {
    bfd_false,
    msp430asm_mkobject,
    _bfd_generic_mkarchive,
    bfd_false,
  },
  {				/* bfd_write_contents */
    bfd_false,
    msp430asm_write_object_contents,
    _bfd_write_archive_contents,
    bfd_false,
  },

  BFD_JUMP_TABLE_GENERIC (msp430asm),
  BFD_JUMP_TABLE_COPY (_bfd_generic),
  BFD_JUMP_TABLE_CORE (_bfd_nocore),
  BFD_JUMP_TABLE_ARCHIVE (_bfd_noarchive),
  BFD_JUMP_TABLE_SYMBOLS (msp430asm),
  BFD_JUMP_TABLE_RELOCS (msp430asm),
  BFD_JUMP_TABLE_WRITE (msp430asm),
  BFD_JUMP_TABLE_LINK (msp430asm),
  BFD_JUMP_TABLE_DYNAMIC (_bfd_nodynamic),

  NULL,
  
  (PTR) 0
};

